#include "caldav.h"

CURL* curl;
char access_token[sizeof(char)*2048] = "";

// Local bind address for receiving OAuth callbacks.
// Reserve 2 chars for the ipv6 square brackets.
char ip[INET6_ADDRSTRLEN], ipstr[INET6_ADDRSTRLEN + 2];

// Static XML data for CalDav post requests:
// https://write.in0rdr.ch/caldav-calendar-discovery
const char* principal_request = "<d:propfind xmlns:d='DAV:' xmlns:cs='http://calendarserver.org/ns/'>"
                                "<d:prop><d:current-user-principal/></d:prop>"
                                "</d:propfind>";
const char* homeset_request = "<d:propfind xmlns:d='DAV:' xmlns:c='urn:ietf:params:xml:ns:caldav'>"
                              " <d:prop>"
                              "  <c:calendar-home-set />"
                              " </d:prop>"
                              "</d:propfind>";
const char* calendar_request = "<d:propfind xmlns:d='DAV:' xmlns:cs='http://calendarserver.org/ns/' xmlns:c='urn:ietf:params:xml:ns:caldav'>"
                               " <d:prop>"
                               "  <d:resourcetype />"
                               "  <d:displayname />"
                               "  <cs:getctag />"
                               "  <c:supported-calendar-component-set />"
                               " </d:prop>"
                               "</d:propfind>";
char* events_request = "<c:calendar-query xmlns:d='DAV:' xmlns:c='urn:ietf:params:xml:ns:caldav'>"
                       "<d:prop><d:getetag/><c:calendar-data/></d:prop>"
                       "<c:filter><c:comp-filter name='VCALENDAR'>"
                       "<c:comp-filter name='VEVENT'>"
                       "<c:time-range start='%s' end='%s'/></c:comp-filter>"
                       "</c:comp-filter></c:filter></c:calendar-query>";

static size_t curl_write_mem_callback(void* contents, size_t size, size_t nmemb, void* userp) {
    size_t realsize = size * nmemb;
    struct curl_mem_chunk* mem = (struct curl_mem_chunk*)userp;

    char* ptr = realloc(mem->memory, mem->size + realsize + 1);
    if (!ptr) {
        tracepoint(diary, error, "Not enough memory (realloc in CURLOPT_WRITEFUNCTION returned NULL)");
        return 0;
    }

    mem->memory = ptr;
    memcpy(&(mem->memory[mem->size]), contents, realsize);
    mem->size += realsize;
    mem->memory[mem->size] = 0;

    return realsize;
}

// todo
// https://beej.us/guide/bgnet/html
void* get_in_addr(struct sockaddr* sa) {
    if (sa->sa_family == AF_INET) {
        return &(((struct sockaddr_in*)sa)->sin_addr);
    }

    return &(((struct sockaddr_in6*)sa)->sin6_addr);
}

/* Get access token from "oauth_eval_cmd" and refresh global var */
void get_access_token() {
    FILE *fp;
    int status;

    fp = popen(CONFIG.oauth_eval_cmd, "r");
    if (fp == NULL)
        perror("Failure running oauth_eval_cmd in get_access_token()");

    while (fgets(access_token, sizeof(char)*2048, fp) != NULL)
        tracepoint(diary, debug_string, "New access token in get_access_token()", access_token);

    // strip any newline characters
    access_token[strcspn(access_token, "\n")] = '\0';

    status = pclose(fp);
    if (status == -1) {
        perror("Failure during pclose in get_access_token()");
    }
}

/* Make a CalDAV request */
char* caldav_req(struct tm* date, char* url, char* http_method, const char* postfields, int depth, bool basicauth, char* content_type, long num_redirects) {
    // only support depths 0 or 1
    if (depth < 0 || depth > 1) {
        return NULL;
    }

    CURLcode res;

    curl = curl_easy_init();

    // https://curl.se/libcurl/c/getinmemory.html
    struct curl_mem_chunk caldav_resp;
    caldav_resp.memory = malloc(1);
    if (caldav_resp.memory == NULL) {
        perror("malloc failed");
        return NULL;
    }
    caldav_resp.size = 0;

    if (curl) {
        // fail if not authenticated, !CURLE_OK
        curl_easy_setopt(curl, CURLOPT_FAILONERROR, 1L);
        //curl_easy_setopt(curl, CURLOPT_VERBOSE, 1);
        curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, http_method);
        tracepoint(diary, debug_string, "curl_easy_perform() request url", url);
        curl_easy_setopt(curl, CURLOPT_URL, url);
        curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, curl_write_mem_callback);
        curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void *)&caldav_resp);

        // construct header fields
        struct curl_slist *header = NULL;

        // prefer basicauth credentials when set
        if (basicauth) {
                char basicauth [strlen(CONFIG.caldav_username) + strlen(CONFIG.caldav_password) + 2];
                sprintf(basicauth, "%s:%s", CONFIG.caldav_username, CONFIG.caldav_password);
                curl_easy_setopt(curl, CURLOPT_USERPWD, basicauth);
        } else {
                // use OAuth credentials
                char bearer_token[strlen("Authorization: Bearer ") + strlen(access_token) + 1];
                sprintf(bearer_token, "Authorization: Bearer %s", access_token);
                header = curl_slist_append(header, bearer_token);
        }

        char depth_header[strlen("Depth: 0") + 1];
        sprintf(depth_header, "Depth: %i", depth);
        header = curl_slist_append(header, depth_header);
        curl_easy_setopt(curl, CURLOPT_HTTPHEADER, header);

        if (strcmp(content_type, "") != 0 && content_type != NULL) {
                char contentType[strlen("Content-Type: ") + strlen(content_type) + 1];
                sprintf(contentType, "Content-Type: %s", content_type);
                header = curl_slist_append(header, contentType);
        }

        // set postfields, if any
        if (postfields != NULL) {
            curl_easy_setopt(curl, CURLOPT_POSTFIELDS, postfields);
        }

        // enable redirect following, some providers have certain caldav
        // endpoinds on sub-paths
        curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 1L);
        curl_easy_setopt(curl, CURLOPT_MAXREDIRS, num_redirects);

        res = curl_easy_perform(curl);

        long response_code = 0;
        curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &response_code);

        // free the memory used for the curl slist
        curl_slist_free_all(header);

        curl_easy_cleanup(curl);

        if (res != CURLE_OK) {
            tracepoint(diary, error_string, "curl_easy_perform() failed in caldav_req()", curl_easy_strerror(res));
            tracepoint(diary, debug_long, "curl_easy_perform() response code", response_code);
            free(caldav_resp.memory);
            return NULL;
        }
    }

    return caldav_resp.memory;
}

/* Upload event to CalDAV server */
void put_event(struct tm* date, const char* dir, size_t dir_size, char* calendar_uri, bool basicauth) {
    // get entry path
    char path[100];
    char* ppath = path;
    char* descr;
    long descr_bytes;

    fpath(dir, dir_size, date, &ppath, sizeof path);
    if (ppath == NULL) {
        tracepoint(diary, error_date, "Cannot get file path for entry", date);
        return;
    }

    FILE* fp = fopen(path, "r");
    if (fp == NULL) perror("Error opening file");

    fseek(fp, 0, SEEK_END);
    descr_bytes = ftell(fp);
    rewind(fp);

    size_t descr_label_size = strlen("DESCRIPTION:");
    size_t descr_size = descr_bytes + descr_label_size;
    descr = calloc(descr_size + 1, sizeof(char));
    if (descr == NULL) {
        perror("malloc failed");
        fclose(fp);
        return;
    }

    strcat(descr, "DESCRIPTION:");

    size_t items_read = fread(descr + descr_label_size, sizeof(char), descr_bytes, fp);
    if (items_read != descr_bytes) {
        perror("Error while reading in put_event()");
        fclose(fp);
        return;
    }

    descr[descr_size] = '\0';
    char* folded_descr = fold(descr);

    char uid[9];
    strftime(uid, sizeof uid, "%Y%m%d", date);

    char* ics = "BEGIN:VCALENDAR\n"
                "PRODID:-//diary\n" // todo: add version
                "VERSION:2.0\n"
                "CALSCALE:GREGORIAN\n"
                "BEGIN:VEVENT\n"
                "UID:%s\n"
                "DTSTART;VALUE=DATE:%s\n"
                "SUMMARY:%s\n"
                "%s\n"
                "END:VEVENT\n"
                "END:VCALENDAR";
    char postfields[strlen(ics) + strlen(folded_descr) + 100];
    sprintf(postfields, ics,
            uid,
            uid,
            uid, // todo: display first few chars of DESCRIPTION as SUMMARY
            folded_descr);

    tracepoint(diary, debug_string, "PUT text/calendar", postfields);

    strcat(calendar_uri, uid);
    strcat(calendar_uri, ".ics");
    char* response = caldav_req(date, calendar_uri, "PUT", postfields, 0, basicauth, "text/calendar", 0);
    fclose(fp);
    free(folded_descr);
    free(descr);

    if (response == NULL) {
        tracepoint(diary, error, "PUT request for event failed in put_event()");
    }

    // free memory allocated to store curl response
    free(response);
}

/*
* Sync with CalDAV server.
* Returns the answer char of the confirmation dialogue
* Returns 0 if neither local nor remote file exists.
* Otherwise, returns -1 on error.
*/
int caldav_sync(struct tm* date,
                 WINDOW* header,
                 WINDOW* cal,
                 int pad_pos,
                 const char* dir,
                 size_t dir_size,
                 bool confirm) {
    pthread_t progress_tid;

    bool oauth_enabled = !(strcmp(CONFIG.oauth_eval_cmd, "") == 0);
    bool basicauth_enabled = !(strcmp(CONFIG.caldav_username, "") == 0 || strcmp(CONFIG.caldav_password, "") == 0);
    tracepoint(diary, debug_int, "oauth_enabled: ", oauth_enabled);
    tracepoint(diary, debug_int, "basicauth_enabled", basicauth_enabled);

    // check remote server and calendar are defined
    if (strcmp(CONFIG.caldav_server, "") == 0 || strcmp(CONFIG.caldav_calendar, "") == 0) {
        tracepoint(diary, debug_string, "CONFIG.caldav_server", CONFIG.caldav_server);
        tracepoint(diary, debug_string, "CONFIG.caldav_calendar", CONFIG.caldav_calendar);
        tracepoint(diary, debug_string, "CONFIG.caldav_username", CONFIG.caldav_username);
        tracepoint(diary, debug_string, "CONFIG.caldav_password", CONFIG.caldav_password);
        show_info(header, "CalDAV config incomplete, press any key to continue.", NULL);
        return -1;
    }

    if (oauth_enabled) {
        if (strcmp(access_token, "") == 0) {
            // get acess token with oauth_eval_cmd
            tracepoint(diary, debug_string, "Fetching access token using oauth_eval_cmd", CONFIG.oauth_eval_cmd);
            get_access_token();
            tracepoint(diary, debug_string, "OAuth access token", access_token);
        }
    }

    pthread_create(&progress_tid, NULL, show_progress, (void*)header);
    pthread_detach(progress_tid);

    char* user_principal = caldav_req(date, CONFIG.caldav_server, "PROPFIND", principal_request, 0, basicauth_enabled, "application/xml", 1);
    tracepoint(diary, debug_string, "User principal return XML", user_principal);

    if (user_principal == NULL) {
        // The principal could not be fetched
        tracepoint(diary, debug, "Unable to fetch user principal");
        show_info(header, "Unable to fetch user principal, press any key to continue.", &progress_tid);
        return -1;
    }

    char* current_user_principal = extract_xml_content(
            user_principal,
            "//*[local-name()='current-user-principal']/*[local-name()='href']",
            header,
            &progress_tid);

    // free memory allocated by curl request
    free(user_principal);

    if (current_user_principal == NULL) {
        tracepoint(diary, debug, "Unable to parse current user principal");
        show_info(header, "Unable to parse current user principal, press any key to continue.", &progress_tid);
        return -1;
    }

    // extract host without path as basis for further api paths
    char *caldav_host; char* caldav_host_scheme;
    CURLU *h = curl_url(); CURLUcode uc = curl_url_set(h, CURLUPART_URL, CONFIG.caldav_server, 0); uc = curl_url_get(h, CURLUPART_HOST, &caldav_host, 0);
    uc = curl_url_get(h, CURLUPART_SCHEME, &caldav_host_scheme, 0);
    tracepoint(diary, debug_string, "Caldav server host name", caldav_host);
    tracepoint(diary, debug_string, "Caldav server scheme/protocol", caldav_host_scheme);
    tracepoint(diary, debug_string, "Caldav server host name", caldav_host);
    if (!uc) {
        tracepoint(diary, error, "cur_url_get() failed in caldav_sync()");
    }
    curl_url_cleanup(h);

    // get the home-set of the user
    char uri[300];
    sprintf(uri, "%s://%s%s", caldav_host_scheme, caldav_host, current_user_principal);

    char* home_set = caldav_req(date, uri, "PROPFIND", homeset_request, 0, basicauth_enabled, "application/xml", 0);
    tracepoint(diary, debug_string, "Home set xml", home_set);

    // parse home set from xml response
    char* home_set_parsed = extract_xml_content(
            home_set,
            "//*[local-name()='calendar-home-set']/*[local-name()='href']",
            header,
            &progress_tid);

    if (home_set_parsed == NULL) {
        tracepoint(diary, debug, "Unable to parse home set");
        show_info(header, "Unable to parse home set, press any key to continue.", &progress_tid);
        return -1;
    }

    sprintf(uri, "%s://%s%s", caldav_host_scheme, caldav_host, home_set_parsed);

    free(home_set);
    free(home_set_parsed);

    // get calendar from home-set
    char* calendar_xml = caldav_req(date, uri, "PROPFIND", calendar_request, 1, basicauth_enabled, "application/xml", 0);
    tracepoint(diary, debug_string, "Calendar set xml", calendar_xml);

    // get calendar URI from the home-set
    char calendar_xpath_query[300];
    sprintf(calendar_xpath_query, "//*[local-name()='displayname'][text()='%s']/../../../*[local-name()='href']", CONFIG.caldav_calendar);
    tracepoint(diary, debug_string, "Calendar xpath query", calendar_xpath_query);
    char* calendar_href = extract_xml_content(
            calendar_xml,
            calendar_xpath_query,
            header,
            &progress_tid);

    if (calendar_href == NULL) {
        tracepoint(diary, debug_string, "Could not find CalDAV calendar", CONFIG.caldav_calendar);
        char* msg_fmtstr = "Could not find CalDAV calendar '%s', press any key to continue.";
        char msg[strlen(msg_fmtstr) + strlen(CONFIG.caldav_calendar)];
        sprintf(msg, msg_fmtstr, CONFIG.caldav_calendar);
        show_info(header, msg, &progress_tid);
        return -1;
    }

    free(calendar_xml);

    char dstr_start[30];
    char dstr_end[30];
    char* format_start = "%Y%m%dT000000Z"; // start of day
    char* format_end = "%Y%m%dT235900Z"; // end of day
    strftime(dstr_start, sizeof dstr_start, format_start, date);
    strftime(dstr_end, sizeof dstr_end, format_end, date);

    char caldata_postfields[strlen(events_request) + strlen(dstr_start) + strlen(dstr_end)];
    sprintf(caldata_postfields, events_request,
            dstr_start,
            dstr_end);

    // fetch event for the cursor date
    sprintf(uri, "%s://%s%s", caldav_host_scheme, caldav_host, calendar_href);

    char* events_xml = caldav_req(date, uri, "REPORT", caldata_postfields, 1, basicauth_enabled, "application/xml", 0);

    if (events_xml == NULL) {
        tracepoint(diary, debug, "Events could not be fetched");
        show_info(header, "Events could not be fetched", &progress_tid);
        return -1;
    }

    // only fetch PRODID diary events
    char* event = extract_xml_content(
            events_xml,
            "//*[local-name()='calendar-data' and contains(text(), 'PRODID:-//diary')]",
            header,
            &progress_tid);

    free(events_xml);

    // no remote event found
    if (event == NULL) {
        tracepoint(diary, debug, "No remote events found, continuing.");
        event = "";
    }

    // get path of entry
    char path[100];
    char* ppath = path;
    fpath(CONFIG.dir, strlen(CONFIG.dir), date, &ppath, sizeof path);

    time_t epoch_zero = 0;

    // assume local file does not exist
    struct tm* localfile_time = gmtime(&epoch_zero);

    // check last modification time of local time
    struct stat attr;
    if (stat(path, &attr) == 0) {
        // if local file does exists, remember utc time of file on disk
        localfile_time = gmtime(&attr.st_mtime);
    }

    time_t localfile_date = mktime(localfile_time);

    struct tm* remote_datetime;
    time_t remote_date = 0;
    long search_pos = 0;

    // check remote DTSTAMP of remote event
    char* remote_dtstamp = extract_ical_field(event, "DTSTAMP", &search_pos, false);
    char* remote_uid = extract_ical_field(event, "UID", &search_pos, false);
    tracepoint(diary, debug_string, "event", event);
    tracepoint(diary, debug_string, "DTSTAMP field", remote_dtstamp);

    // init remote date to 1970, assume remote file does not exist
    remote_datetime = gmtime(&epoch_zero);

    if (remote_dtstamp != NULL) {
        // if remote date does exist, set date to the correct timestamp
        strptime(remote_dtstamp, "%Y%m%dT%H%M%SZ", remote_datetime);
        free(remote_dtstamp);
    }

    remote_date = mktime(remote_datetime);

    tracepoint(diary, debug_string, "Remote last modified", ctime(&remote_date));
    tracepoint(diary, debug_string, "Local last modified", ctime(&localfile_date));
    double timediff = difftime(localfile_date, remote_date);
    tracepoint(diary, debug_double, "Time diff between local and remote mod time", timediff);

    char* rmt_desc;
    char dstr[16];
    int conf_ch = 0;

    if (timediff == 0) {
        tracepoint(diary, debug, "Local and remote files have equal timestamp or don't exist, giving up.");
        pthread_cancel(progress_tid);
        wclear(header);
        // free memory allocated to store curl response
        curl_free(caldav_host_scheme);
        curl_free(caldav_host);
        free(remote_uid);
        return 0;
    } else if (timediff > 0) {
        tracepoint(diary, debug, "Local file is newer. Uploading to remote.");
        if (remote_uid) {
            // purge any existing daily calendar entries on the remote side
            char event_uri[300];
            sprintf(event_uri, "%s://%s%s%s.ics", caldav_host_scheme, caldav_host, calendar_href, remote_uid);
            //sprintf(event_uri, "%s%s%s.ics", CONFIG.caldav_server, calendar_href, remote_uid);
            tracepoint(diary, debug_string, "Event URI for DELETE request", event_uri);
            char* response = caldav_req(date, event_uri, "DELETE", NULL, 0, basicauth_enabled, "", 0);
            free(response);
        }

        put_event(date, dir, dir_size, uri, basicauth_enabled);

        pthread_cancel(progress_tid);
        wclear(header);

    } else if (timediff < 0) {
        rmt_desc = extract_ical_field(event, "DESCRIPTION", &search_pos, true);

        if (rmt_desc == NULL) {
            tracepoint(diary, error, "Could not fetch description of remote event. Aborting sync.");
            pthread_cancel(progress_tid);
            wclear(header);
            curl_free(caldav_host_scheme);
            curl_free(caldav_host);
            free(remote_uid);
            return -1;
        }

        if (confirm) {
            // prepare header for confirmation dialogue
            curs_set(2);
            noecho();
            pthread_cancel(progress_tid);
            wclear(header);

            // ask for confirmation
            strftime(dstr, sizeof dstr, CONFIG.fmt, date);
            mvwprintw(header, 0, 0, "Remote event is more recent. Sync entry '%s' and overwrite local file? [(Y)es/(a)ll/(n)o/(c)ancel] ", dstr);
            conf_ch = wgetch(header);
        }

        if (conf_ch == 'y' || conf_ch == 'Y' || conf_ch == 'a' || conf_ch == '\n' || !confirm) {
            tracepoint(diary, debug, "Remote file is newer. Extracting description from remote.");
            char* i;

            // persist downloaded buffer to local file
            FILE* cursordate_file = fopen(path, "wb");
            if (cursordate_file == NULL) {
                perror("Failed to open cursor date file");
            } else {
                for (i = rmt_desc; *i != '\0'; i++) {
                    if (rmt_desc[i-rmt_desc] == 0x5C) { // backslash
                        switch (*(i+1)) {
                            case 'n':
                                fputc('\n', cursordate_file);
                                i++;
                                break;
                            case 0x5c: // preserve real backslash
                                fputc(0x5c, cursordate_file);
                                i++;
                                break;
                        }
                    } else {
                        fputc(*i, cursordate_file);
                    }
                }
            }
            fclose(cursordate_file);

            // add new entry highlight
            chtype atrs = winch(cal) & A_ATTRIBUTES;
            wchgat(cal, 2, atrs | A_BOLD, 0, NULL);
            prefresh(cal, pad_pos, 0, 1, ASIDE_WIDTH, LINES - 1, ASIDE_WIDTH + CAL_WIDTH);
        }

        echo();
        curs_set(0);
        free(rmt_desc);
    }

    // free memory allocated to store curl response
    curl_free(caldav_host_scheme);
    curl_free(caldav_host);
    free(remote_uid);
    free(calendar_href);
    pthread_join(progress_tid, NULL);
    wclear(header);
    return conf_ch;
}
